unit Graph2D;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, StdCtrls, ExtCtrls, Math, Controls, Windows, Dialogs, Forms, Graphics;                         // Dialogs löschen!

type
 TGraph2D = class;
 TGraph2DMovePoint = class(TButton)
  private
   _X,_Y      : Double;
   Move       : Boolean;
   StartC     : TPoint;
   StartB     : TPoint;
   procedure MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
   procedure MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
   procedure MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
  public
   OnMove : TNotifyEvent;
   constructor Create(_Owner : TGraph2D; X,Y : Double);
   property X : Double read _X write _X;
   property Y : Double read _Y write _Y;
   procedure Free;
 end;
 TGraph2DPoint = packed record
  X : Double;
  Y : Double;
 end;
 TGraph2DSpecialPoint = packed record
  X : Double;
  Y : Double;
  t : Byte;
 end;
 TRenderMethod = procedure(Graph : TGraph2D);
 TRenderMethodObj = procedure(Graph : TGraph2D) of Object;

 TGraph2DLine = Array of TGraph2DPoint;
 TGraph2D = class(TPanel)
  private
   ViewRX1 : Double;
   ViewRY1 : Double;
   ViewRX2 : Double;
   ViewRY2 : Double;

   KeyR    : Boolean;
   KeyL    : Boolean;
   KeyT    : Boolean;
   KeyB    : Boolean;
   ShiftS  : TShiftState;
   Timer   : TTimer;

   Move    : Boolean;
   StartC  : TPoint;
   StartPX : Double;
   StartPY : Double;

   KeyCatcher : TEdit;
   procedure DrawAxis;
   procedure DrawLines;
   procedure DrawMovePoints;
   procedure MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
   procedure MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
   procedure MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
   procedure MouseWheel(Sender: TObject; Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
   procedure TimerEV(Sender : TObject);
   procedure KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
   procedure KeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
  public
   Lines      : Array of TGraph2DLine;
   RenderFkts : Array of TMethod;
   MovePoints : Array of TGraph2DMovePoint;
   OnMove     : TNotifyEvent;
   constructor Create(_Owner : TComponent);
   procedure SetViewArea(X1,Y1,X2,Y2 : Double);
   function XPosToScreen(Pos : Double) : integer;
   function YPosToScreen(Pos : Double) : integer;
   function XScreenToPos(Screen : integer) : Double;
   function YScreenToPos(Screen : integer) : Double;
   procedure DoPaint(Sender : TObject);
   procedure Free;
 end;

implementation

const
 MinRange = 0.000000001;
 MaxRange = 1000000000;
 PixDiffPerLine = 50;
 Speed = 10;

constructor TGraph2D.Create(_Owner : TComponent);
begin
 inherited Create(_Owner);
 Self.OnPaint := @DoPaint;
 SetViewArea(-5,5,5,-5);
 Move := false;
 OnMove := nil;
 OnMouseDown := @MouseDown;
 OnMouseUp := @MouseUp;
 OnMouseMove := @MouseMove;
 OnMouseWheel := @MouseWheel;
 KeyL := false;
 KeyR := false;
 KeyT := false;
 KeyB := false;
 ShiftS := [];
 Timer := TTimer.Create(self);
 Timer.Interval := 40;
 Timer.Enabled := false;
 Timer.OnTimer := @TimerEv;
 OnMove := nil;
 OnMouseDown := @MouseDown;
 OnMouseUp := @MouseUp;
 OnMouseMove := @MouseMove;
 OnMouseWheel := @MouseWheel;
 OnKeyDown := @KeyDown;
 OnKeyUp := @KeyUp;
 KeyCatcher := TEdit.Create(self);
 KeyCatcher.Parent := self;
 KeyCatcher.OnKeyDown := @KeyDown;
 KeyCatcher.OnKeyUp := @KeyUp;
 KeyCatcher.ReadOnly := true;
 KeyCatcher.OnEnter := @DoPaint;
 KeyCatcher.OnExit := @DoPaint;
 KeyCatcher.Top := -100;
end;

procedure TGraph2D.SetViewArea(X1,Y1,X2,Y2 : Double);
var
 Cash : Double;
begin
 if (X2 < X1) then begin Cash := X1; X1 := X2; X2 := Cash; end;
 if (Y2 > Y1) then begin Cash := Y1; Y1 := Y2; Y2 := Cash; end;
 if (X2 - X1 < MinRange) then X2 := X1 + MinRange;
 if (Y1 - Y2 < MinRange) then Y2 := Y1 - MinRange;
 if (X2 - X1 > MaxRange) then X2 := X1 + MaxRange;
 if (Y1 - Y2 > MaxRange) then Y2 := Y1 - MaxRange;
 ViewRX1 := X1;
 ViewRY1 := Y1;
 ViewRX2 := X2;
 ViewRY2 := Y2;
 if (OnMove <> nil) then OnMove(self);
end;

procedure TGraph2D.MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 KeyCatcher.SetFocus;
 StartC.X := X;
 StartC.Y := Y;
 StartPX := ViewRX1;
 StartPY := ViewRY1;
 Move := true;
end;

procedure TGraph2D.MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 Move := false;
end;

procedure TGraph2D.MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var d,e : Double;
begin
 if (Move) then begin
  d := StartPX + (StartC.X - X) / Width * (ViewRX2 - ViewRX1);
  e := StartPY + (StartC.Y - Y) / Height * (ViewRY2 - ViewRY1);
  SetViewArea(d,e,d + ViewRX2 - ViewRX1,e + ViewRY2 - ViewRY1);
  DoPaint(Sender);
 end;
end;

procedure TGraph2D.MouseWheel(Sender: TObject; Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
var
 Sx,Sy                   : Single;
 WheelDeltaX,WheelDeltaY : integer;
begin
 if (Width <= 0) then exit;
 if (Height <= 0) then exit;
 WheelDeltaX := Round(WheelDelta / 1000 * Width);
 WheelDeltaY := Round(WheelDelta / 1000 * Height);
 Sx := MousePos.X / Width;
 Sy := MousePos.Y / Height;
 SetViewArea(XScreenToPos(Round(WheelDeltaX * Sx)),YScreenToPos(Round(WheelDeltaY * Sy)),XScreenToPos(Width - Round(WheelDeltaX * (1 - Sx))),YScreenToPos(Height - Round(WheelDeltaY * (1 - Sy))));
 DoPaint(Sender);
end;

procedure TGraph2D.TimerEV(Sender : TObject);
var SX1,SY1,SX2,SY2 : shortint;
begin
 if (not KeyT) and (not KeyB) and (not KeyL) and (not KeyR) then begin Timer.Enabled := false; exit; end;
 SX1 := 0;
 SY1 := 0;
 SX2 := 0;
 SY2 := 0;
 if (KeyR) and (not KeyL) then begin SX1 := Speed; SX2 := Speed; end;
 if (KeyL) and (not KeyR) then begin SX1 := -Speed; SX2 := -Speed; end;
 if (KeyR) and (KeyL) and (ssCtrl in ShiftS) then begin SX1 := -Speed; SX2 := Speed; end;
 if (KeyR) and (KeyL) and (not (ssCtrl in ShiftS)) then begin SX1 := Speed; SX2 := -Speed; end;

 if (KeyB) and (not KeyT) then begin SY1 := Speed; SY2 := Speed; end;
 if (KeyT) and (not KeyB) then begin SY1 := -Speed; SY2 := -Speed; end;
 if (KeyT) and (KeyB) and (ssCtrl in ShiftS) then begin SY1 := -Speed; SY2 := Speed; end;
 if (KeyT) and (KeyB) and (not (ssCtrl in ShiftS)) then begin SY1 := Speed; SY2 := -Speed; end;

 SetViewArea(XScreenToPos(SX1),YScreenToPos(SY1),XScreenToPos(Width + SX2),YScreenToPos(Height + SY2));
 DoPaint(Sender);
end;

procedure TGraph2D.KeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
 if (Key = 37) then KeyL := true
 else if (Key = 38) then KeyT := true
 else if (Key = 39) then KeyR := true
 else if (Key = 40) then KeyB := true;
 ShiftS := Shift;
 Timer.Enabled := true;
end;

procedure TGraph2D.KeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
 if (Key = 37) then KeyL := false
 else if (Key = 38) then KeyT := false
 else if (Key = 39) then KeyR := false
 else if (Key = 40) then KeyB := false;
 ShiftS := Shift;
end;

function TGraph2D.XPosToScreen(Pos : Double) : integer;
var D : Double;
begin
 D := ((Pos - ViewRX1) / (ViewRX2 - ViewRX1)) * Width;
 if (D > 1000000) then Result := 1000000
 else if (D < -1000000) then Result := -1000000
 else Result := Round(D);
end;

function TGraph2D.YPosToScreen(Pos : Double) : integer;
var D : Double;
begin
 D := (Pos - ViewRY1) / (ViewRY2 - ViewRY1) * Height;
 if (D > 1000000) then Result := 1000000
 else if (D < -1000000) then Result := -1000000
 else Result := Round(D);
end;

function TGraph2D.XScreenToPos(Screen : integer) : Double;
begin
 Result := 0;
 if (Width <= 0) then exit;
 Result := ViewRX1 + (ViewRX2 - ViewRX1) * (Screen / Width);
end;

function TGraph2D.YScreenToPos(Screen : integer) : Double;
begin
 Result := 0;
 if (Height <= 0) then exit;
 Result := ViewRY1 + (ViewRY2 - ViewRY1) * (Screen / Height);
end;

function IntPower(Basis : Double; Exp : Integer) : Double;
var i : integer;
begin
 Result := 1;
 for i := 0 to abs(Exp) - 1 do Result := Result * Basis;
 if (Exp < 0) then Result := 1 / Result;
end;

procedure GetPosOfRegularPoints(From,UpTo : Double; Count : Word; var Start,Space : Double);
var
 Exp : integer;
 K   : Byte;
 D   : Double;
function GetNextSpace : Double;
begin
 if (K = 5) then K := 2
 else if (K = 2) then K := 1
 else begin
  K := 5;
  dec(Exp);
 end;
 Result := K * IntPower(10,Exp);
end;
begin
 Exp := floor(Ln(Abs(UpTo - From)) / Ln(10)) + 1;
 K := 2;
 Space := K * IntPower(10,Exp);
 repeat begin
  D := GetNextSpace;
  if (D * Count >= UpTo - From) then begin
   Space := D;
   Start := ceil(From / Space) * Space;
  end;
 end until (D * Count < UpTo - From);
 // showmessage('From: ' + floattostr(From) + ' To: ' + floattostr(UpTo) + ' Count: ' + inttostr(Count) + ' Start: ' + floattostr(Start) + ' Space: ' + floattostr(Space));
end;

procedure TGraph2D.DrawAxis;
var
 Pos        : integer;
 LineLength : integer;
 i,j,k      : integer;
 d,e        : double;
begin
 Canvas.Pen.Color := clBlack;
 Pos := XPosToScreen(0);
 if (Pos < 0) then Pos := 0;
 if (Pos >= Width) then Pos := Width - 1;
 k := Pos + 1;
 if (k + 20 > Width) then k := Width - 20;
 LineLength := Floor(Height / PixDiffPerLine);
 GetPosOfRegularPoints(ViewRY2,ViewRY1,LineLength,d,e);
 for i := 0 to LineLength - 1 do begin
  j := YPosToScreen(d + e*i);
  Canvas.TextOut(k,j - 1,FloatToStr(d + e*i));
  Canvas.Line(Pos - 2,j,Pos + 2,j);
 end;
 Canvas.Line(Pos,0,Pos,Height);

 Pos := YPosToScreen(0);
 if (Pos < 0) then Pos := 0;
 if (Pos >= Height) then Pos := Height - 1;
 k := Pos + 1;
 if (k + 18 > Height) then k := Height - 18;
 LineLength := Floor(Width / PixDiffPerLine);
 GetPosOfRegularPoints(ViewRX1,ViewRX2,LineLength,d,e);
 for i := 0 to LineLength - 1 do begin
  j := XPosToScreen(d + e*i);
  Canvas.TextOut(j + 1,k,FloatToStr(d + e*i));
  Canvas.Line(j,Pos + 2,j,Pos - 2);
 end;
 Canvas.Line(0,Pos,Width,Pos);
end;

procedure TGraph2D.DrawLines;
var i,j : integer;
begin
 for i := 0 to Length(Lines) - 1 do if (Length(Lines[i]) > 0) then begin
  Canvas.MoveTo(XPosToScreen(Lines[i][0].X),YPosToScreen(Lines[i][0].Y));
  for j := 1 to Length(Lines[i]) - 1 do Canvas.LineTo(XPosToScreen(Lines[i][j].X),YPosToScreen(Lines[i][j].Y));
 end;
end;

procedure TGraph2D.DrawMovePoints;
var i : integer;
begin
 for i := 0 to Length(MovePoints) - 1 do begin
  MovePoints[i].Left := XPosToScreen(MovePoints[i].X) - MovePoints[i].Width div 2;
  MovePoints[i].Top := YPosToScreen(MovePoints[i].Y) - MovePoints[i].Height div 2;
 end;
end;

procedure TGraph2D.DoPaint(Sender : TObject);
var i : integer;
begin
 Canvas.Brush.Color := clInfoBk;
 Canvas.FillRect(0,0,Width,Height);
 DrawAxis;
 DrawLines;
 for i := 0 to Length(RenderFkts)-1 do if (RenderFkts[i].Data = nil) then TRenderMethod(RenderFkts[i].Code)(self) else TRenderMethodObj(RenderFkts[i])(self);
 DrawMovePoints;
 if (KeyCatcher.Focused) then begin
  Canvas.Brush.Color := clBlue;
  Canvas.FrameRect(0,0,Width,Height);
 end;
end;

procedure TGraph2D.Free;
begin
 inherited Free;
end;





constructor TGraph2DMovePoint.Create(_Owner : TGraph2D; X,Y : Double);
begin
 if (not Assigned(_Owner)) then exit;
 inherited Create(_Owner);
 Caption := '';
 Width := 15;
 Height := 15;
 Left := _Owner.XPosToScreen(X) - Width div 2;
 Top := _Owner.YPosToScreen(Y) - Height div 2;
 _X := X;
 _Y := Y;
 Move := false;
 OnMove := nil;
 Parent := _Owner;
 SetLength(_Owner.MovePoints,Length(_Owner.MovePoints) + 1);
 _Owner.MovePoints[Length(_Owner.MovePoints) - 1] := self;
 OnMouseDown := @MouseDown;
 OnMouseUp := @MouseUp;
 OnMouseMove := @MouseMove;
end;

procedure TGraph2DMovePoint.MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 GetCursorPos(StartC);
 StartB.X := Left;
 StartB.Y := Top;
 Move := true;
end;

procedure TGraph2DMovePoint.MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
 Move := false;
end;

procedure TGraph2DMovePoint.MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
 CurPos : TPoint;
begin
 if (Move) then begin
  GetCursorPos(CurPos);
  X := StartB.X + CurPos.X - StartC.X;
  Y := StartB.Y + CurPos.Y - StartC.Y;
  _X := TGraph2D(Owner).XScreenToPos(X);
  _Y := TGraph2D(Owner).YScreenToPos(Y);
  if (OnMove = nil) then begin Left := X; Top := Y; end
  else OnMove(self);
  TGraph2D(Owner).DoPaint(Sender);
 end;
end;

procedure TGraph2DMovePoint.Free;
var i : integer;
begin
 if (Length(TGraph2D(Owner).MovePoints) > 0) then begin
  for i := 0 to Length(TGraph2D(Owner).MovePoints) - 1 do if (TGraph2D(Owner).MovePoints[i] = self) then break;
  if (TGraph2D(Owner).MovePoints[i] = self) then begin
   for i := i to Length(TGraph2D(Owner).MovePoints) - 2 do TGraph2D(Owner).MovePoints[i] := TGraph2D(Owner).MovePoints[i + 1];
   SetLength(TGraph2D(Owner).MovePoints,Length(TGraph2D(Owner).MovePoints) - 1);
  end;
 end;
 inherited Free;
end;

end.

