unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls,
  ExtCtrls, PairSplitter, Buttons, ComCtrls, Unit2, StrTransformErr, Graph2D,
  PolynomUtil, DataClasses, xmlUtil;

type

 TFormGeschwMessung = class;

 TGeschwindigkeitsMessung = class(TDataItem)
  type
   TMeasurePoint = record
     Time     : Single;   // Sekunden nach Start
     Velocity : Single;   // m/s
     Power    : Single;   // w
     Name     : String;   // Name des Wegpunktes
     Way      : Single;   // Weg in m von 1. Wegpunkt
     Alt      : Single;   // Höhe des Wegpunktes über NN
   end;

  private
   Start : TDateTime;
   Data  : Array of TMeasurePoint;
   Form  : TFormGeschwMessung;
   function ReadPoints(i : Cardinal) : TMeasurePoint; inline;
   procedure WritePoints(i : Cardinal; P : TMeasurePoint); inline;
   function ReadCount : Cardinal; inline;
   procedure WriteCount(c : Cardinal); inline;
   function ToStrings : TStrings;
   procedure FromStrings(St : TStrings);
   procedure updateInfo;

  protected
   destructor Destroy; override;

  public
   DayOfWeek : integer;
   Hour      : Single;
   constructor Create;
   property Points[i : Cardinal] : TMeasurePoint read ReadPoints write WritePoints; default;
   property Count : Cardinal read ReadCount write WriteCount;
   procedure Edit(Sender : TObject); override;
   function DataToXML : XML.TElement; override;
   procedure XMLToData(_Data : XML.TElement); override;
 end;

 { TFormGeschwMessung }

 TFormGeschwMessung = class(TForm)
   Button1: TButton;
   Button2: TButton;
   Button3: TButton;
   Button5: TButton;
   Button6: TButton;
   Button7: TButton;
   CheckBox1: TCheckBox;
   Edit1: TEdit;
   Label1: TLabel;
   Label2: TLabel;
   Label3: TLabel;
   Label4: TLabel;
   LabelMesspunkt: TLabel;
   Memo1: TMemo;
   PairSplitter1: TPairSplitter;
   PairSplitter2: TPairSplitter;
   PairSplitterSide1: TPairSplitterSide;
   PairSplitterSide2: TPairSplitterSide;
   GraphW: TGraph2D;
   GraphV: TGraph2D;
   PairSplitterSide3: TPairSplitterSide;
   PairSplitterSide4: TPairSplitterSide;
   Panel1: TPanel;
   TrackBar1: TTrackBar;
   procedure Add(Sender: TObject; var Key: Word; Shift: TShiftState);
   procedure AdjustViewRect(Sender: TObject);
   procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
   procedure FormCreate(Sender: TObject);
   procedure FormResize(Sender: TObject);
   procedure KompensateChanged(Sender: TObject);
   procedure MessungChanged(Sender: TObject);
   procedure ReachPos(Sender: TObject);
   procedure Start(Sender: TObject);
   procedure Stopp(Sender: TObject);
   procedure OpenEnergyConsumption(Sender: TObject);
   procedure TextToMessung(Sender: TObject);
   constructor Create(M : TGeschwindigkeitsMessung);

  type
   Poly = specialize Polynom<Extended>;
   PolyS = specialize Polynom<Single>;

  protected
   destructor Destroy; override;

  private
   Pos     : integer;
   Fw      : Poly.IFunction;
   Fv      : Poly.IFunction;
   Messung : TGeschwindigkeitsMessung;
   procedure RenderPoints(Graph : TGraph2D);
   procedure RenderGraphs(Graph : TGraph2D);
   procedure CheckViewRect;
   procedure AddValue(v : TGeschwindigkeitsMessung.TMeasurePoint);
   procedure CalcF;
 end;


implementation

type
 Poly = TFormGeschwMessung.Poly;
 PolyS = TFormGeschwMessung.PolyS;

const
  Tab : String = Char(9);

{$R *.lfm}



constructor TGeschwindigkeitsMessung.Create;
begin
 inherited;
 SetLength(Data, 0);
 Form := nil;
 updateInfo;
end;

procedure TGeschwindigkeitsMessung.updateInfo;
var
  h, m, s, ms : Word;
begin
 if (Length(Data) = 0) then begin
  Hour := 0;
  DayOfWeek := 0;
  if (Form <> nil) then Form.Caption := Title;
  exit;
 end;
 DayOfWeek := sysutils.DayOfWeek(Start);
 sysutils.DecodeTime(Start, h, m, s, ms);
 Hour := h + m/60 + s/3600;
 Title := sysutils.FormatDateTime('YYYY.MM.DD hh:nn', Start);
 if (Form <> nil) then Form.Caption := Title;
end;

function TGeschwindigkeitsMessung.ReadPoints(i : Cardinal) : TMeasurePoint; inline;
begin
 Result := Data[i];
end;

procedure TGeschwindigkeitsMessung.WritePoints(i : Cardinal; P : TMeasurePoint); inline;
begin
 Data[i] := P;
end;

function TGeschwindigkeitsMessung.ReadCount : Cardinal; inline;
begin
 Result := Length(Data);
end;

procedure TGeschwindigkeitsMessung.WriteCount(c : Cardinal); inline;
begin
 SetLength(Data, c);
end;

procedure TGeschwindigkeitsMessung.Edit(Sender : TObject);
begin
 if (Form = nil) then Form := TFormGeschwMessung.Create(self);
 Form.Show;
 Form.Memo1.Lines := ToStrings;
 Form.CheckViewRect;
 Form.CalcF;
 Form.Button5.Enabled := false;
end;

procedure TGeschwindigkeitsMessung.FromStrings(St : TStrings);
var
  i,j,k,c : integer;
  SelPos : Array [0..3] of integer;
  P : TGeschwindigkeitsMessung.TMeasurePoint;
  s : String;
  M : Array of TGeschwindigkeitsMessung.TMeasurePoint;
begin
 c := 0;
 SetLength(M, 0);
 if (Form <> nil) and (Form.Memo1.Lines <> St) then Form.Memo1.Lines := St;
 for i := 0 to St.Count - 1 do begin
  if (St[i] = '') then continue;
  P.Time := -1;
  P.Velocity := PolyS.NaN;
  P.Power := PolyS.NaN;
  P.Name := '';
  P.Way := PolyS.NaN;
  P.Alt := PolyS.NaN;
  k := 0;
  s := '';
  for j := 1 to Length(St[i]) + 1 do if (St[i][j] = Tab) or (Byte(St[i][j]) = 0) then begin
   case k of
    0 : P.Time := StrToFloatErr(s, -1);
    1 : P.Velocity := StrToFloatErr(s, PolyS.NaN);
    2 : P.Power := StrToFloatErr(s, PolyS.NaN);
    3 : P.Name := s;
    4 : P.Way := StrToFloatErr(s, PolyS.NaN);
    5 : P.Alt := StrToFloatErr(s, PolyS.NaN);
   end;
   if (k <= 5) then SelPos[k] := j;
   s := '';
   k := k + 1;
  end else s := s + St[i][j];
  if (P.Time < 0) and (Form <> nil) then begin
   Form.memo1.selstart := c;
   Form.Memo1.SelLength := SelPos[0]-1;
   Form.Memo1.SetFocus;
   if (MessageDlg('Ungültiger Messpunkt', 'Die Zeile ' + IntToStr(i) + '("' + St[i] + '") hat keinen gültigen Zeitstempel. Soll diese Zeile übergangen werden?', mtConfirmation, [mbIgnore, mbAbort],0) = mrIgnore) then continue else exit;
  end else if (P.Time < 0) then continue;
  c := c + Length(St[i]) + 2; // Absatzgröße hier!
  k := Length(M);
  SetLength(M, k+1);
  M[k] := P;
 end;
 Data := M;
 if (Form <> nil) then begin
  Form.CheckViewRect;
  Form.CalcF;
  Form.Button5.Enabled := false;
 end;
 updateInfo;
 ShowMessage('FromStrings ist fertig!');
end;

function TGeschwindigkeitsMessung.ToStrings : TStrings;
var
  i : integer;
  s : String;
begin
 Result := TStringList.Create;
 for i := 0 to Length(Data) - 1 do begin
  s := FloatToStr(Data[i].Time) + Tab;
  if (not TFormGeschwMessung.PolyS.IsNaN(Data[i].Velocity)) then s := s + FloatToStr(Data[i].Velocity);
  s := s + Tab;
  if (not TFormGeschwMessung.PolyS.IsNaN(Data[i].Power)) then s := s + FloatToStr(Data[i].Power);
  s := s + Tab + Data[i].Name + Tab;
  if (not TFormGeschwMessung.PolyS.IsNaN(Data[i].Way)) then s := s + FloatToStr(Data[i].Way);
  s := s + Tab;
  if (not TFormGeschwMessung.PolyS.IsNaN(Data[i].Alt)) then s := s + FloatToStr(Data[i].Alt);
  Result.add(s);
 end;
end;

function TGeschwindigkeitsMessung.DataToXML : XML.TElement;
begin
 Result := XML.TElement.Create;
 Result.tagName := 'MESSUNG';
 Result.setAttribute('title', title);
 Result.setAttribute('start', FloatToStr(Start));
 Result.appendChild(XML.TText.Create(ToStrings));
end;

procedure TGeschwindigkeitsMessung.XMLToData(_Data : XML.TElement);
begin
 SetLength(Data, 0);
 if (_Data.tagName <> 'MESSUNG') then exit;
 title := _Data.GetAttribute('title');
 Start := StrToFloatErr(_Data.GetAttribute('start'), 0);
 FromStrings(_Data.getTextContent(true));    // Fehler Hier!
 ShowMessage('XMLToData ist fertig!');
end;

destructor TGeschwindigkeitsMessung.Destroy;
begin
 if (Form <> nil) then Form.Free;
 SetLength(Data, 0);
 inherited;
end;





constructor TFormGeschwMessung.Create(M : TGeschwindigkeitsMessung);
begin
 inherited Create(nil);
 Fw := nil;
 Fv := nil;
 Messung := M;
 Messung.Form := self;
 Caption := Messung.Title;
end;

procedure TFormGeschwMessung.FormCreate(Sender: TObject);
begin
 GraphW := TGraph2D.Create(self);
 GraphW.Parent := PairSplitterSide3;
 GraphW.AnchorParallel(akTop, 0, PairSplitterSide3);
 GraphW.AnchorParallel(akLeft, 0, PairSplitterSide3);
 GraphW.AnchorParallel(akRight, 0, PairSplitterSide3);
 GraphW.AnchorParallel(akBottom, 0, PairSplitterSide3);
 SetLength(GraphW.RenderFkts, 2);
 GraphW.RenderFkts[0] := TMethod(@RenderPoints);
 GraphW.RenderFkts[1] := TMethod(@RenderGraphs);
 GraphV := TGraph2D.Create(self);
 GraphV.Parent := PairSplitterSide4;
 GraphV.AnchorParallel(akTop, 0, PairSplitterSide4);
 GraphV.AnchorParallel(akLeft, 0, PairSplitterSide4);
 GraphV.AnchorParallel(akRight, 0, PairSplitterSide4);
 GraphV.AnchorParallel(akBottom, 0, PairSplitterSide4);
 SetLength(GraphV.RenderFkts, 2);
 GraphV.RenderFkts[0] := TMethod(@RenderPoints);
 GraphV.RenderFkts[1] := TMethod(@RenderGraphs);
end;

procedure TFormGeschwMessung.Start(Sender: TObject);
begin
 if (Route.Count = 0) then begin
  ShowMessage('Messung kann nicht gestartet werden, da keine Route definiert wurde!');
  exit;
 end;
 Memo1.Lines.Clear;
 SetLength(Messung.Data, 0);
 Messung.Start := 0;
 Pos := 0;
 Button1.enabled := false;
 Button2.enabled := true;
 Button3.enabled := true;
 Edit1.enabled := true;
 if (CheckBox1.Checked) then LabelMesspunkt.Caption := Route.Reverse[Pos].Name
 else LabelMesspunkt.Caption := Route[Pos].Name;
 Memo1.ReadOnly := true;
 Button5.Enabled := false;
 CheckBox1.Enabled := false;
end;

procedure TFormGeschwMessung.ReachPos(Sender: TObject);
var v : TGeschwindigkeitsMessung.TMeasurePoint;
begin
 if (Pos >= Route.Count) then exit;
 v.Velocity := StrToFloatErr(Edit1.Text, PolyS.NaN);
 if (not PolyS.IsNaN(v.Velocity)) then v.Velocity := v.Velocity * 0.2777777777777777;
 if (Messung.Start = 0) then begin Messung.Start := Now; v.Time := 0; end
 else v.Time := (Now-Messung.Start)*100000;
 if (CheckBox1.Checked) then begin
  v.Name := Route.Reverse[Pos].Name;
  v.Way := Route.Reverse[Pos].Way;
  v.Alt := Route.Reverse[Pos].Alt;
 end else begin
  v.Name := Route[Pos].Name;
  v.Way := Route[Pos].Way;
  v.Alt := Route[Pos].Alt;
 end;
 v.Power := PolyS.NaN;
 AddValue(v);
 Edit1.Text := '';
 Edit1.SetFocus;
 Pos := Pos + 1;
 if (Pos >= Route.Count) then Stopp(Sender)
 else if (CheckBox1.Checked) then LabelMesspunkt.Caption := Route.Reverse[Pos].Name
 else LabelMesspunkt.Caption := Route[Pos].Name;
end;

procedure TFormGeschwMessung.Add(Sender: TObject; var Key: Word; Shift: TShiftState);
var
  v : TGeschwindigkeitsMessung.TMeasurePoint;
  t : Double;
begin
 if (Key <> 13) or (Messung.Count = 0) then exit;
 v.Velocity := StrToFloatErr(Edit1.Text, PolyS.NaN);
 if (PolyS.IsNaN(v.Velocity)) then begin
  Edit1.SelectAll;
  exit;
 end;
 v.Velocity := v.Velocity * 0.2777777777777777;
 v.Power := PolyS.NaN;
 if (Messung.Start = 0) then begin Messung.Start := Now; v.Time := 0; end
 else v.Time := (Now-Messung.Start)*100000;
 v.Name := '';
 v.Way := PolyS.NaN;
 v.Alt := PolyS.NaN;
 AddValue(v);
 Edit1.Text := '';
end;

procedure TFormGeschwMessung.AdjustViewRect(Sender: TObject);
begin
 CheckViewRect;
end;

procedure TFormGeschwMessung.Stopp(Sender: TObject);
begin
  Button1.enabled := true;
  Button2.enabled := false;
  Button3.enabled := false;
  Edit1.enabled := false;
  CheckBox1.Enabled := true;
  LabelMesspunkt.Caption := '';
  Edit1.Text := '';
  Memo1.Lines := Messung.ToStrings;
  Memo1.ReadOnly := false;
  CalcF;
end;

procedure TFormGeschwMessung.OpenEnergyConsumption(Sender: TObject);
var
  p : Poly.TPolynom;
  m : Poly.TMatrix;
begin
  p := Poly.TPolynom.Create;
  p[0] := 1;
  p[1] := 3;
  p[2] := 2;
  ShowMessage(p.ToString);
  m.Init(4, 3, false);
  m[3,2] := 5.32;
  ShowMessage(FloatToStr(m[3,2]));
end;

procedure TFormGeschwMessung.TextToMessung(Sender: TObject);
begin
 Messung.FromStrings(Memo1.Lines);
end;

procedure TFormGeschwMessung.FormResize(Sender: TObject);
begin
  if (self.Width < 900) then self.Width := 900;
end;

procedure TFormGeschwMessung.KompensateChanged(Sender: TObject);
begin
  CalcF;
end;

procedure TFormGeschwMessung.MessungChanged(Sender: TObject);
begin
  if (Button1.enabled) then Button5.Enabled := true;
end;


procedure TFormGeschwMessung.RenderPoints(Graph : TGraph2D);
var
  i, x, y : integer;
  P : Array [0..2] of TPoint;
begin
  if (Messung.Count = 0) then exit;
  Graph.Canvas.Brush.Color := clBlack;
  for i := 0 to Messung.Count-1 do begin
   x := Graph.XPosToScreen(Messung[i].Time);
   if (Graph = self.GraphW) then begin
    if (not PolyS.IsNaN(Messung[i].Way)) then begin
     y := Graph.YPosToScreen(Messung[i].Way);
     Graph.Canvas.FillRect(x-4, y-4, x+4, y+4);
    end;
    if (not PolyS.IsNaN(Messung[i].Alt)) then begin
     Graph.Canvas.Pen.Color := clBlue;
     y := Graph.YPosToScreen(Messung[i].Alt);
     Graph.Canvas.FillRect(x-4, y-4, x+4, y+4);
    end;
   end else if (Graph = self.GraphV) then begin
    if (not PolyS.IsNaN(Messung[i].Velocity)) then begin
     y := Graph.YPosToScreen(Messung[i].Velocity);
     P[0].x := x-4;  P[0].y := y+4;
     P[1].x := x+4;  P[1].y := y+4;
     P[2].x := x;  P[2].y := y-4;
     Graph.Canvas.Polygon(P);
    end;
    if (not PolyS.IsNaN(Messung[i].Power)) then begin
     y := Graph.YPosToScreen(Messung[i].Power);
     Graph.Canvas.Ellipse(x-4, y-4, x+4, y+4);
    end;
    if (not PolyS.IsNaN(Messung[i].Way)) then begin
     Graph.Canvas.MoveTo(x, 0);
     Graph.Canvas.LineTo(x, Graph.height);
    end;
   end;
  end;
end;

procedure TFormGeschwMessung.RenderGraphs(Graph : TGraph2D);
var
  f     : Poly.Precision;
  b     : Boolean;
  x, y  : integer;
  P     : Poly.IFunction;
begin
 if (Graph = self.GraphW) then P := Fw else P := Fv;
 if (P = nil) then exit;
 b := true;
 for x := 0 to Graph.Canvas.Width - 1 do begin
  f := P.Calc(Graph.XScreenToPos(x));
  if (Poly.IsNaN(f)) then begin b := true; continue; end;
  y := Graph.YPosToScreen(f);
  if (b) then begin Graph.Canvas.MoveTo(x, y); b := false; end
  else Graph.Canvas.LineTo(x, y);
 end;
end;

procedure TFormGeschwMessung.CheckViewRect;
var
 i : integer;
 t : double;
 minX, maxX, minYw, maxYw, minYv, maxYv : double;

function lt(A, B : Single) : Boolean;
begin
 if (PolyS.IsNaN(B)) then Result := true
 else if (PolyS.IsNaN(A)) then Result := false
 else Result := A < B;
end;
function gt(A, B : Single) : Boolean;
begin
 if (PolyS.IsNaN(B)) then Result := true
 else if (PolyS.IsNaN(A)) then Result := false
 else Result := A > B;
end;

begin
 minX := 0;
 maxX := 1;
 minYw := 0;
 maxYw := 1;
 minYv := 0;
 maxYv := 1;
 for i := 0 to Messung.Count - 1 do begin
  t := Messung[i].Time;
  if (lt(t, minX)) then minX := t;
  if (gt(t, maxX)) then maxX := t;
  if (lt(Messung[i].Way, minYw)) then minYw := Messung[i].Way;
  if (gt(Messung[i].Way, maxYw)) then maxYw := Messung[i].Way;
  if (lt(Messung[i].Velocity, minYv)) then minYv := Messung[i].Velocity;
  if (gt(Messung[i].Velocity, maxYv)) then maxYv := Messung[i].Velocity;
 end;
 t := (maxX - minX)*0.05;
 maxX := maxX + t;
 minX := minX - t;
 t := (maxYw - minYw)*0.05;
 maxYw := maxYw + t;
 minYw := minYw - t;
 t := (maxYv - minYv)*0.05;
 maxYv := maxYv + t;
 minYv := minYv - t;
 self.GraphW.SetViewArea(minX, minYw, maxX, maxYw);
 self.GraphW.DoPaint(nil);
 self.GraphV.SetViewArea(minX, minYv, maxX, maxYv);
 self.GraphV.DoPaint(nil);
end;

procedure TFormGeschwMessung.AddValue(v : TGeschwindigkeitsMessung.TMeasurePoint);
var
  l : integer;
  s : String;
begin
 l := Messung.Count;
 Messung.Count := l+1;
 Messung[l] := v;
 s := FloatToStr(v.Time) + Tab;
 if (not PolyS.IsNaN(v.Velocity)) then s := s + FloatToStr(v.Velocity);
 s := s + Tab;
 if (not PolyS.IsNaN(v.Power)) then s := s + FloatToStr(v.Power);
 s := s + Tab + v.Name + Tab;
 if (not PolyS.IsNaN(v.Way)) then s := s + FloatToStr(v.Way);
 s := s + Tab;
 if (not PolyS.IsNaN(v.Alt)) then s := s + FloatToStr(v.Alt);
 self.Memo1.Lines.Add(s);
 CheckViewRect;
 if (l = 0) then Messung.updateInfo;
end;

procedure TFormGeschwMessung.CalcF;
var
  IP  : Array of Poly.TInterpolatePoint;
  i,j : integer;
begin
 if (not self.Button1.Enabled) then exit;
 SetLength(IP, Messung.Count * 2);
 j := -1;
 for i := Messung.Count - 1 downto 0 do begin
  if (not PolyS.IsNaN(Messung[i].Way)) then begin
   j := j + 1;
   IP[j].x := Messung[i].Time;
   IP[j].y := Messung[i].Way;
   IP[j].derivation := 0;
  end;
  if (not PolyS.IsNaN(Messung[i].Velocity)) then begin
   j := j + 1;
   IP[j].x := Messung[i].Time;
   IP[j].y := Messung[i].Velocity;
   IP[j].derivation := 1;
  end;
 end;
 SetLength(IP, j+1);
 Fw := Poly.AdjustedSplinesOnFirstDerivation(IP, self.TrackBar1.Position / self.TrackBar1.Max * 5);
 Fv := Fw.Derivate;
 self.GraphW.DoPaint(nil);
 self.GraphV.DoPaint(nil);
end;


procedure TFormGeschwMessung.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  if (Fw <> nil) then Fw.Free; Fw := nil;
  if (Fv <> nil) then Fv.Free; Fv := nil;
end;

destructor TFormGeschwMessung.Destroy;
begin
 Messung.Form := nil;
 inherited;
end;

end.

