unit Unit2;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls, StrTransformErr, xmlUtil;

type
  TFormRoute = class;

  TRoute = class
   type
    TRoutePoint = record
     Name : String;
     Way  : Single;
     Alt  : Single;
    end;

   private
    _Route : Array of TRoutePoint;
    _xml   : XML.TElement;
    Form   : TFormRoute;
    function ReadRoute(i : Cardinal) : TRoutePoint; inline;
    function ReadReverse(i : Cardinal) : TRoutePoint;
    function ReadCount : Cardinal; inline;
    procedure SetXML(el : XML.TElement);

   protected
    destructor Destroy; override;

   public
    constructor Create;
    property Route[i : Cardinal] : TRoutePoint read ReadRoute; default;
    property Reverse[i : Cardinal] : TRoutePoint read ReadReverse;
    property Count : Cardinal read ReadCount;
    procedure FromStrings(Str : TStrings);
    function ToStrings : TStrings;
    property XMLElement : XML.TElement read _xml write SetXML;
    procedure Edit(Sender : TObject);
  end;

  { TFormRoute }

  TFormRoute = class(TForm)
    Button1: TButton;
    Label1: TLabel;
    Memo1: TMemo;
    procedure FormCreate(Sender: TObject);
    procedure SetRoute(Sender: TObject);
  private
    Route : TRoute;
  protected
    destructor Destroy; override;
  public
    constructor Create(R : TRoute);
  end;

var
 Route : TRoute;

implementation

{$R *.lfm}

constructor TRoute.Create;
begin
 inherited;
 Form := nil;
end;

procedure TRoute.Edit(Sender : TObject);
begin
 if (Form = nil) then Form := TFormRoute.Create(self);
 Form.Show;
end;

function TRoute.ReadRoute(i : Cardinal) : TRoutePoint; inline;
begin
 Result := _Route[i];
end;

function TRoute.ReadReverse(i : Cardinal) : TRoutePoint;
var l : Cardinal;
begin
 l := Length(_Route)-1;
 Result := _Route[l-i];
 Result.Way := _Route[l].Way - Result.Way;
end;

function TRoute.ReadCount : Cardinal; inline;
begin
 Result := Length(_Route);
end;

procedure TRoute.FromStrings(Str : TStrings);
var
  R : Array of TRoutePoint;
  P : TRoutePoint;
  i,j,k : integer;
  b : Byte;
  s : String;
  NaN : Single;
begin
 Cardinal(NaN) := $FFFFFFFF;
 SetLength(R, 0);
 for i := 0 to Str.Count - 1 do begin
  k := Length(Str[i])+1;
  if (k < 3) then continue;
  s := '';
  b := 0;
  P.Name := '';
  P.Way := NaN;
  P.Alt := NaN;
  for j := 1 to k+1 do if (Byte(Str[i][j]) = 9) or (Byte(Str[i][j]) = 0) then begin
   case (b) of
    0 : P.Name := s;
    1 : P.Way := StrToFloatErr(s, NaN);
    2 : P.Alt := StrToFloatErr(s, NaN);
   end;
   s := '';
   b := b + 1;
  end else s := s + Str[i][j];
  if (Cardinal(P.Way) = Cardinal(NaN)) or (P.Way < 0) then begin ShowMessage('Fehler: Der Wegpunkt "' + P.Name + '" hat eine ungültige Streckenangabe!'); exit; end;
  if (Cardinal(P.Alt) = Cardinal(NaN)) then begin ShowMessage('Fehler: Der Wegpunkt "' + P.Name + '" hat eine ungültige Höhenangabe!'); exit; end;
  j := Length(R);
  if (j > 0) and (P.Way <= R[j-1].Way) then begin ShowMessage('Fehler: Der Wegpunkt "' + P.Name + '" hat eine kleinere Streckenangabe, als sein Vorgänger!'); exit; end;
  SetLength(R, j+1);
  R[j] := P;
 end;
 _Route := R;
 if (Form <> nil) and (Form.Memo1.Lines <> Str) then Form.Memo1.Lines := Str;
end;

function TRoute.ToStrings : TStrings;
var
  i : integer;
begin
 Result := TStringList.Create;
 for i := 0 to Length(_Route)-1 do begin
  Result.Add(_Route[i].Name + Char(9) + FloatToStr(_Route[i].Way) + Char(9) + FloatToStr(_Route[i].Alt));
 end;
end;

procedure TRoute.SetXML(el : XML.TElement);
begin
 if (Form <> nil) then Form.Free;
 if (_xml <> nil) then begin
  _xml.clearNodes;
  _xml.appendChild(XML.TText.Create(ToStrings));
  _xml.tagName := 'ROUTE';
 end;
 _xml := el;
 if (_xml <> nil) then FromStrings(_xml.getTextContent(true));
end;

destructor TRoute.Destroy;
begin
 if (Form <> nil) then Form.Free;
 inherited;
end;



constructor TFormRoute.Create(R : TRoute);
begin
 inherited Create(nil);
 Route := R;
end;

procedure TFormRoute.FormCreate(Sender: TObject);
begin
 Memo1.Lines := Route.ToStrings;
end;

procedure TFormRoute.SetRoute(Sender: TObject);
begin
 Route.FromStrings(Memo1.Lines);
end;

destructor TFormRoute.Destroy;
begin
 Route.Form := nil;
 inherited;
end;

initialization
Route := TRoute.Create;

end.

