unit unitStartForm;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, Menus,
  StdCtrls, ExtCtrls, PairSplitter, DataClasses, Unit1, xmlUtil, Unit2, unitZaehlung;

type

  { TForm3 }

  TForm3 = class(TForm)
    Button1: TButton;
    Button2: TButton;
    Button3: TButton;
    Edit1: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    MenuItem5: TMenuItem;
    OpenDialog1: TOpenDialog;
    PairSplitter1: TPairSplitter;
    PairSplitterSide1: TPairSplitterSide;
    PairSplitterSide2: TPairSplitterSide;
    Panel1: TPanel;
    Panel2: TPanel;
    Panel3: TPanel;
    Panel4: TPanel;
    Routen: TListBox;
    MainMenu1: TMainMenu;
    MenuItem1: TMenuItem;
    MenuItem2: TMenuItem;
    MenuItem3: TMenuItem;
    MenuItem4: TMenuItem;
    SaveDialog1: TSaveDialog;
    Messungen: TDataClass;
    Zaehlungen: TDataClass;
    procedure ChooseItem(Sender: TObject);
    procedure DeleteRoute(Sender: TObject);
    procedure EditRoute(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure NewFile(Sender: TObject);
    procedure NewRoute(Sender: TObject);
    procedure OpenFile(Sender: TObject);
    procedure RouteTitleChanged(Sender: TObject);
    procedure Save(Sender: TObject);
    procedure SaveAs(Sender: TObject);
  private
    function createItemG(tagName : ShortString) : TDataItem;
    function createItemZ(tagName : ShortString) : TDataItem;
  public
    { public declarations }
  end;

var
  Form3     : TForm3;
  LastFile  : String = '';
  Modified  : Boolean = false;

implementation

{$R *.lfm}

{ TForm3 }

procedure TForm3.FormCreate(Sender: TObject);
begin
 Messungen := TDataClass.Create(self);
 Messungen.title := 'Geschwindigkeitsmessungen:';
 Messungen.Parent := Panel3;
 Messungen.onCreateNode := @createItemG;
 Messungen.AnchorParallel(akTop, 0, Panel3);
 Messungen.AnchorParallel(akLeft, 0, Panel3);
 Messungen.AnchorParallel(akRight, 0, Panel3);
 Messungen.AnchorParallel(akBottom, 0, Panel3);
 Zaehlungen := TDataClass.Create(self);
 Zaehlungen.title := 'Verkehrszählungen:';
 Zaehlungen.Parent := Panel4;
 Zaehlungen.onCreateNode := @createItemZ;
 Zaehlungen.AnchorParallel(akTop, 0, Panel4);
 Zaehlungen.AnchorParallel(akLeft, 0, Panel4);
 Zaehlungen.AnchorParallel(akRight, 0, Panel4);
 Zaehlungen.AnchorParallel(akBottom, 0, Panel4);
end;

procedure TForm3.NewFile(Sender: TObject);
begin
 Routen.ItemIndex := -1;
 Routen.Clear;
 LastFile := '';
end;

function TForm3.createItemG(tagName : ShortString) : TDataItem;
begin
 Result := TGeschwindigkeitsMessung.Create;
end;

function TForm3.createItemZ(tagName : ShortString) : TDataItem;
begin
 Result := TVerkehrszaehlung.Create;
end;

procedure TForm3.ChooseItem(Sender: TObject);
var
 RI : XML.TElement;
function Get(t : ShortString) : XML.TElement;
begin
 if (RI = nil) then begin Result := nil; exit; end;
 Result := RI.getElementsByTagName(t, 0);
 if (Result = nil) then begin
  Result := XML.TElement.Create;
  Result.tagName := t;
  RI.appendChild(Result);
 end;
end;
begin
 if (Routen.ItemIndex = -1) then begin
  RI := nil;
  Edit1.Text := '';
  Button2.Enabled := false;
  Edit1.Enabled := false;
 end else begin
  RI := XML.TElement(Routen.Items.Objects[Routen.ItemIndex]);
  Edit1.Text := RI.GetAttribute('title');
  Button2.Enabled := true;
  Edit1.Enabled := true;
 end;
 Route.XMLElement := Get('ROUTE');
 Messungen.XMLElement := Get('MESSUNGEN');
 Zaehlungen.XMLElement := Get('ZAEHLUNGEN');
 Modified := true;
end;

procedure TForm3.DeleteRoute(Sender: TObject);
begin
  if (Routen.ItemIndex = -1) then exit;
  if (MessageDlg('Löschen', 'Sind Sie sicher, dass Sie die gewählte Route samt Messungen löschen möchten?', mtConfirmation, [mbYes, mbNo], 0) = mrNo) then exit;
  Routen.Items.Objects[Routen.ItemIndex].Free;
  Routen.Items.Delete(Routen.ItemIndex);
  ChooseItem(nil);
end;

procedure TForm3.EditRoute(Sender: TObject);
begin
 if (Route = nil) then exit;
 Route.Edit(nil);
end;

procedure TForm3.NewRoute(Sender: TObject);
var
  RI : XML.TElement;
  R : XML.TElement;
  M : XML.TElement;
  Z : XML.TElement;
begin
 RI := XML.TElement.Create;
 RI.tagName := 'ROUTEDATA';
 RI.setAttribute('title', 'NeueRoute ' + IntToStr(Routen.Count));
 R := XML.TElement.Create;
 R.tagName := 'ROUTE';
 RI.appendChild(R);
 M := XML.TElement.Create;
 M.tagName := 'MESSUNGEN';
 RI.appendChild(M);
 Z := XML.TElement.Create;
 Z.tagName := 'ZAEHLUNGEN';
 RI.appendChild(Z);
 Routen.AddItem(RI.GetAttribute('title'), RI);
end;

procedure TForm3.OpenFile(Sender: TObject);
var
  a : Array of XML.TNode;
  f : TStringList;
  i : integer;
begin
 if (not OpenDialog1.Execute) then exit;
 LastFile := OpenDialog1.FileName;
 Routen.ItemIndex := -1;
 Routen.Clear;
 f := TStringList.Create;
 f.LoadFromFile(OpenDialog1.FileName);
 a := XML.FromStrings(f);
 f.Free;
 for i := 0 to Length(a)-1 do if (a[i].nodeType = ELEMENT_NODE) then begin
  Routen.AddItem(XML.TElement(a[i]).getAttribute('title'), XML.TElement(a[i]));
 end;
 SetLength(a, 0);
end;

procedure TForm3.RouteTitleChanged(Sender: TObject);
var i : integer;
begin
 i := Routen.ItemIndex;
 if (i = -1) then exit;
 Routen.Items[i] := Edit1.Text;
 XML.TElement(Routen.Items.Objects[i]).SetAttribute('title', Edit1.Text);
end;

procedure TForm3.Save(Sender: TObject);
var
  t : TStringList;
  i : integer;

procedure Append(s : TStrings); inline;
var j : integer;
begin
 for j := 0 to s.Count - 1 do t.add(s[j]);
end;

begin
 Routen.ItemIndex := -1; ChooseItem(nil);
 if (LastFile = '') then begin ShowMessage('Kein Dateiname angegeben!'); exit; end;
 t := TStringList.Create;
 for i := 0 to Routen.Count - 1 do Append(XML.ToStrings(XML.TElement(Routen.Items.Objects[i])));
 try
  t.SaveToFile(LastFile);
  Modified := false;
 finally
  t.Free;
 end;
end;

procedure TForm3.SaveAs(Sender: TObject);
begin
 if (not SaveDialog1.Execute) then exit;
 if (FileExists(SaveDialog1.FileName)) and (MessageDlg('Überschreiben', 'Die gewählte Datei existiert bereits. Überschreiben?', mtConfirmation, [mbYes, mbNo], 0) = mrNo) then exit;
 LastFile := SaveDialog1.FileName;
 Save(Sender);
end;

procedure TForm3.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
 if (Modified) and (MessageDlg('Schließen', 'Möglicher Weise liegen ungesicherte Änderungen vor. Trotzdem Schließen?', mtConfirmation, [mbYes, mbAbort], 0) = mrAbort) then CloseAction := caNone;
end;

end.

