unit unitZaehlung;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls,
  ExtCtrls, xmlUtil, DataClasses, Windows, StrTransformErr;

type
  TVerkehrszaehlungForm = class;

  TVerkehrszaehlung = class(TDataItem)
   type
    TZaehlung = packed record
     Motorrad     : Array [0..1] of Cardinal;
     PKW          : Array [0..1] of Cardinal;
     Jeep         : Array [0..1] of Cardinal;
     Transporter  : Array [0..1] of Cardinal;
     LKW          : Array [0..1] of Cardinal;
    end;
   private
    Form : TVerkehrszaehlungForm;
    function ReadCount : Cardinal;
   protected
    destructor Destroy; override;
   public
    Data  : TZaehlung;
    Start : TDateTime;
    Stop  : TDateTime;
    constructor Create;
    function ToStrings : TStrings;
    procedure FromStrings(st : TStrings);
    procedure Edit(Sender : TObject); override;
    function DataToXML : XML.TElement; override;
    procedure XMLToData(_Data : XML.TElement); override;
    property Count : Cardinal read ReadCount;
  end;

  { TVerkehrszaehlungForm }

  TVerkehrszaehlungForm = class(TForm)
    ButRichtung: TButton;
    ButPKWL: TButton;
    ButJeepL: TButton;
    ButTransporterL: TButton;
    ButLKWL: TButton;
    ButMotoradR: TButton;
    ButPKWR: TButton;
    ButJeepR: TButton;
    ButTransporterR: TButton;
    ButLKWR: TButton;
    Button1: TButton;
    Button2: TButton;
    ButMotoradL: TButton;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    LabelGes: TLabel;
    Label7: TLabel;
    Memo1: TMemo;
    Panel1: TPanel;
    procedure Add(Sender: TObject);
    procedure SetData(Sender: TObject);
    procedure StartStop(Sender: TObject);
    procedure TurnDirection(Sender: TObject);
  private
   Zaehlung : TVerkehrszaehlung;
   Richtung : Byte;
   Started  : Boolean;
  protected
   destructor Destroy; override;
  public
   constructor Create(_Zaehlung : TVerkehrszaehlung);
  end;

var
  VerkehrszaehlungForm: TVerkehrszaehlungForm;

implementation

{$R *.lfm}

constructor TVerkehrszaehlung.Create;
begin
 inherited;
 Form := nil;
 FillDWord(Data, SizeOf(Data) div 4, 0);
 Start := 0;
 Stop := 0;
end;

procedure TVerkehrszaehlung.Edit(Sender : TObject);
begin
 if (Form = nil) then Form := TVerkehrszaehlungForm.Create(self);
 Form.Show;
end;

function TVerkehrszaehlung.ReadCount : Cardinal;
begin
 Result :=  Data.Motorrad[0] + Data.Motorrad[1]
          + Data.PKW[0] + Data.PKW[1]
          + Data.Jeep[0] + Data.Jeep[1]
          + Data.Transporter[0] + Data.Transporter[1]
          + Data.LKW[0] + Data.LKW[1];
end;

function TVerkehrszaehlung.ToStrings : TStrings;
begin
 Result := TStringList.Create;
 Result.Add('Motorrad' + Char(9) + IntToStr(Data.Motorrad[0]) + Char(9) + IntToStr(Data.Motorrad[1]));
 Result.Add('PKW' + Char(9) + IntToStr(Data.PKW[0]) + Char(9) + IntToStr(Data.PKW[1]));
 Result.Add('Jeep' + Char(9) + IntToStr(Data.Jeep[0]) + Char(9) + IntToStr(Data.Jeep[1]));
 Result.Add('Transporter' + Char(9) + IntToStr(Data.Transporter[0]) + Char(9) + IntToStr(Data.Transporter[1]));
 Result.Add('LKW' + Char(9) + IntToStr(Data.LKW[0]) + Char(9) + IntToStr(Data.LKW[1]));
end;

procedure TVerkehrszaehlung.FromStrings(St : TStrings);
var
  i,j,k,c : integer;
  SelPos : Array [0..3] of integer;
  typ : String;
  n, nRev : Cardinal;
  s : String;
  Z : TZaehlung;
procedure nextVal; inline;
begin
 try
  case k of
   0 : typ := s;
   1 : n := StrToInt(s);
   2 : nRev := StrToInt(s);
  end;
 except
   typ := '';
 end;
 if (k <= 4) then SelPos[k] := j;
 s := '';
 k := k + 1;
end;
type
 TArr2 = Array [0..1] of Cardinal;
function Arr2(a, b : Cardinal) : TArr2; inline;
begin
 Result[0] := a;
 Result[1] := b;
end;
begin
 c := 0;
 FillDWord(Z, SizeOf(Z) div 4, 0);
 if (Form <> nil) and (Form.Memo1.Lines <> St) then Form.Memo1.Lines := St;
 for i := 0 to St.Count - 1 do begin
  if (St[i] = '') then continue;
  typ := '';
  n := 0;
  nRev := 0;
  k := 0;
  s := '';
  for j := 1 to Length(St[i]) do if (St[i][j] = Char(9)) then nextVal
  else s := s + St[i][j];
  nextVal;
  c := c + Length(St[i]) + 2; // Absatzgröße hier!
  case LowerCase(typ) of
   'motorrad' : Z.Motorrad := Arr2(n, nRev);
   'pkw' : Z.PKW := Arr2(n, nRev);
   'jeep' : Z.Jeep := Arr2(n, nRev);
   'transporter' : Z.Transporter := Arr2(n, nRev);
   'lkw' : Z.LKW := Arr2(n, nRev);
  end;
 end;
 Data := Z;
end;

function TVerkehrszaehlung.DataToXML : XML.TElement;
begin
 if (Form <> nil) and (Form.Started) then Form.StartStop(nil);
 Result := XML.TElement.Create;
 Result.tagName := 'ZAEHLUNG';
 Result.setAttribute('title', title);
 Result.setAttribute('start', FloatToStr(Start));
 Result.setAttribute('stop', FloatToStr(Stop));
 Result.appendChild(XML.TText.Create(ToStrings));
end;

procedure TVerkehrszaehlung.XMLToData(_Data : XML.TElement);
begin
 FillDWord(Data, SizeOf(Data) div 4, 0);
 if (_Data.tagName <> 'ZAEHLUNG') then exit;
 title := _Data.GetAttribute('title');
 start := StrToFloatErr(_Data.GetAttribute('start'), 0);
 stop := StrToFloatErr(_Data.GetAttribute('stop'), 0);
 FromStrings(_Data.getTextContent(true));
end;

destructor TVerkehrszaehlung.Destroy;
begin
 if (Form <> nil) then Form.Free;
 inherited;
end;



constructor TVerkehrszaehlungForm.Create(_Zaehlung : TVerkehrszaehlung);
var tabWidth : Integer;
begin
 inherited Create(nil);
 Zaehlung := _Zaehlung;
 Richtung := 0;
 Started := false;
 TabWidth := 50;
 SendMessage(Memo1.Handle, EM_SETTABSTOPS, 1, Longint(@TabWidth));
 Memo1.Lines := Zaehlung.ToStrings;
 LabelGes.Caption := IntToStr(Zaehlung.Count);
end;

procedure TVerkehrszaehlungForm.TurnDirection(Sender: TObject);
begin
 if (Richtung = 0) then Richtung := 1 else Richtung := 0;
 if (Richtung = 0) then ButRichtung.Caption := '←' else ButRichtung.Caption := '→';
end;

procedure TVerkehrszaehlungForm.Add(Sender: TObject);
var
 p : PCardinal;
function dir(b:Byte) : byte;
begin
 if (b = Richtung) then Result := 0 else Result := 1;
end;
begin
 if (not Started) then exit;
 p := nil;
 if (Sender = ButMotoradL) then p := @Zaehlung.Data.Motorrad[dir(0)]
 else if (Sender = ButPKWL) then p := @Zaehlung.Data.PKW[dir(0)]
 else if (Sender = ButJeepL) then p := @Zaehlung.Data.Jeep[dir(0)]
 else if (Sender = ButTransporterL) then p := @Zaehlung.Data.Transporter[dir(0)]
 else if (Sender = ButLKWL) then p := @Zaehlung.Data.LKW[dir(0)]
 else if (Sender = ButMotoradR) then p := @Zaehlung.Data.Motorrad[dir(1)]
 else if (Sender = ButPKWR) then p := @Zaehlung.Data.PKW[dir(1)]
 else if (Sender = ButJeepR) then p := @Zaehlung.Data.Jeep[dir(1)]
 else if (Sender = ButTransporterR) then p := @Zaehlung.Data.Transporter[dir(1)]
 else if (Sender = ButLKWR) then p := @Zaehlung.Data.LKW[dir(1)];
 if (p = nil) then exit;
 p^ := p^ + 1;
 Memo1.Lines := Zaehlung.ToStrings;
 LabelGes.Caption := IntToStr(Zaehlung.Count);
end;

procedure TVerkehrszaehlungForm.SetData(Sender: TObject);
begin
  Zaehlung.FromStrings(Memo1.Lines);
  LabelGes.Caption := IntToStr(Zaehlung.Count);
end;

procedure TVerkehrszaehlungForm.StartStop(Sender: TObject);
begin
 if (Started) then begin
  Started := false;
  Button1.Caption := 'Neue Messung Starten';
  Zaehlung.Stop := Now;
  Memo1.ReadOnly := false;
 end else begin
  Started := true;
  Button1.Caption := 'Messung Stoppen';
  Zaehlung.Start := Now;
  Zaehlung.Title := sysutils.FormatDateTime('YYYY.MM.DD hh:nn', Zaehlung.Start);
  Zaehlung.Stop := 0;
  FillDWord(Zaehlung.Data, SizeOf(Zaehlung.Data) div 4, 0);
  Memo1.ReadOnly := true;
 end;
 Memo1.Lines := Zaehlung.ToStrings;
end;

destructor TVerkehrszaehlungForm.Destroy;
begin
 Zaehlung.Form := nil;
 inherited;
end;

end.

