unit xmlUtil;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, dialogs;

type
 XML = class

  type
   TNodeType = (UNKNOWN_NODE = 0, ELEMENT_NODE = 1, TEXT_NODE = 3);
   TElement = class;

   TNode = class
    private
     _nodeType : TNodeType;
     _parentEl : TElement;
    protected
     destructor Destroy; override;
    public
     property nodeType : TNodeType read _nodeType;
     property parentElement : TElement read _parentEl;
   end;

   TAttribute = object
    public
     name  : ShortString;
     value : String;
   end;
   PAttribute = ^TAttribute;

   TText = class(TNode)
    public
     text : TStrings;
     constructor Create (_text : TStrings);
   end;

   TElement = class(TNode)
    type
     TChildNodes = object
      private
       n : Array of TNode;
       function ReadNodes(i : Cardinal) : TNode; inline;
       function ReadCount : Cardinal; inline;
      public
       property nodes[i : Cardinal] : TNode read ReadNodes; default;
       property count : Cardinal read ReadCount;
     end;
     PChildNodes = ^TChildNodes;
     TAttributes = object
      private
       a : Array of TAttribute;
       function ReadNodes(i : Cardinal) : TAttribute; inline;
       function ReadCount : Cardinal; inline;
      public
       property nodes[i : Cardinal] : TAttribute read ReadNodes; default;
       property count : Cardinal read ReadCount;
     end;

    protected
     destructor Destroy; override;

    public
     tagName    : ShortString;
     childNodes : TChildNodes;
     attributes : TAttributes;
     constructor Create;
     procedure appendChild(newChild : TNode);
     procedure removeChild(i : Cardinal); overload;
     procedure removeChild(child : TNode); overload;
     function getAttribute(name : ShortString) : String;
     procedure setAttribute(name : ShortString; value : String);
     function getElementsByTagName(_tagName : ShortString; i : Cardinal) : TElement;
     function getTextContent(RemoveSpaces : Boolean) : TStrings;
     procedure clearNodes;
   end;

   TNodes = Array of TNode;

 public
  class function ToStrings(rootEl : TElement) : TStrings;
  class function FromStrings(str : TStrings) : TNodes;
 end;

implementation

class function XML.FromStrings(str : TStrings) : TNodes;
var
  i, j, k : integer;
  el, pel : XML.TElement;
  tx : XML.TText;
  c : Byte;
  State : Byte;
  s, t : String;
procedure Next; inline;
label NextStart;
begin
 NextStart:
 if (i >= str.Count) then begin c := 0; exit; end;
 if (j > Length(str[i])) then begin j := 1; i := i + 1; c := 13; exit; end;
 c := Byte(str[i][j]);
 j := j + 1;
 if (c = 10) then goto NextStart;
end;
begin
 SetLength(Result, 0);
 i := 0;
 j := 1;
 el := nil;
 tx := nil;
 State := 0;
 Next;
 while (c <> 0) do begin
  case State of
   0 : begin  // kein Status, von hier wird entweder nach 1 oder nach 2 gesprungen
    if (c = Byte('<')) then begin State := 2; Next; end
    else if (c <> 32) and (c <> 9) and (c <> 13) then begin
     State := 1;
     tx := TText.Create(TStringList.Create);
     if (tx.text.Count = 0) then tx.text.add('');
     k := Length(Result);
     if (el = nil) then begin SetLength(Result, k+1); Result[k] := tx; end else el.appendChild(tx);
    end else Next;
   end;
   1 : begin    // innerhalb einer Text-Node, springt nach 0 zurück
    if (c = Byte('<')) then State := 0
    else begin
     k := tx.text.Count-1;
     if (c = 13) then tx.text.add('')
     else tx.text[k] := tx.text[k] + Char(c);
     Next;
    end;
   end;
   2 : begin    // innerhalb eines XML-Begins, erstellt entweder ein neues Element und springt nach 3 oder springt nach 6
    if (c = 32) or (c = 13) or (c = 9) then Next
    else if (c = Byte('/')) then begin Next; s := ''; State := 6; end
    else if (c = Byte('>')) then begin Next; State := 0; end else begin
     State := 3;
     s := '';
     pel := el;
     el := XML.TElement.Create;
     if (pel = nil) then begin
      k := Length(Result);
      SetLength(Result, k+1);
      Result[k] := el;
     end else pel.appendChild(el);
    end;
   end;
   3 : begin   // TagName lesen, springt nach 0 oder nach 4
    if (c = 32) or (c = 13) or (c = 9) or (c = Byte('>')) then begin
     el.tagName := s;
     s := '';
     if (c = Byte('>')) then begin Next; State := 0; end
     else State := 4;
    end else begin s := s + Char(c); Next; end;
   end;
   4 : begin   // Attribut-Name lesen
    if (c = 32) or (c = 13) or (c = 9) then begin end
    else if (c = Byte('>')) then begin State := 0; end
    else if (c = Byte('=')) or (c = Byte('"')) then begin
     if (c = Byte('=')) and (s <> '') then begin
      Next;
      if (c <> Byte('"')) then ShowMessage('" erwartet, aber ' + Char(c) + ' found (Zeile ' + IntToStr(i) + ', Spalte ' + IntToStr(j) + ')!');
      t := '';
      State := 5;
     end else ShowMessage('Ungültiges Zeichen entdeckt (Zeile ' + IntToStr(i) + ', Spalte ' + IntToStr(j) + ')!');
    end else s := s + Char(c);
    Next;
   end;
   5 : begin   // Attribut-Value lesen
    if (c = Byte('"')) then begin
     el.setAttribute(s, t);
     s := '';
     State := 4;
    end else t := t + Char(c);
    Next;
   end;
   6 : begin   // XML-End-Tag
    if (c = 32) or (c = 13) or (c = 9) then begin end
    else if (c = Byte('>')) then begin
     if (el = nil) then ShowMessage('End-Tag "' + s + '" gefunden, der zu keinem Element gehört (Zeile ' + IntToStr(i) + ', Spalte ' + IntToStr(j) + ')!')
     else begin
      if (el.tagName <> s) then ShowMessage('End-Tag "</' + el.tagName + '>" erwartet, aber "</' + s + '>" gefunden (Zeile ' + IntToStr(i) + ', Spalte ' + IntToStr(j) + ')!');
      el := el.parentElement;
     end;
     State := 0;
    end else s := s + Char(c);
    Next;
   end;
  end;
 end
end;

class function XML.ToStrings(rootEl : TElement) : TStrings;
function doShift(shift : Integer) : AnsiString; inline;
begin
 SetLength(Result, shift);
 FillChar(Result[1], shift, ' ');
end;
procedure NodeToStr(n : TElement; shift : Integer);
var
  i, j : integer;
  s    : AnsiString;
begin
 with n do begin
  s := doShift(shift) + '<' + tagName;
  for i := 0 to Length(attributes.a) - 1 do s := s + ' ' + attributes.a[i].name + '="' + attributes.a[i].value + '"';
  s := s + '>';
  Result.add(s);
  for i := 0 to Length(childNodes.n) - 1 do case childNodes.n[i]._nodeType of
   ELEMENT_NODE : NodeToStr(TElement(childNodes.n[i]), shift + 1);
   TEXT_NODE : for j := 0 to TText(childNodes.n[i]).text.Count - 1 do Result.add(doShift(shift+1) + TText(childNodes.n[i]).text[j]);
  end;
  s := doShift(shift) + '</' + tagName + '>';
  Result.add(s);
 end;
end;
begin
 Result := TStringList.Create;
 NodeToStr(rootEl, 0);
end;


constructor XML.TText.Create(_text : TStrings);
begin
 inherited Create;
 _nodeType := TEXT_NODE;
 _parentEl := nil;
 text := _text;
end;

constructor XML.TElement.Create;
begin
 inherited;
 _nodeType := ELEMENT_NODE;
 _parentEl := nil;
 tagName := '';
 SetLength(childNodes.n, 0);
 SetLength(attributes.a, 0);
end;

function XML.TElement.TChildNodes.ReadNodes(i : Cardinal) : TNode; inline;
begin
 Result := n[i];
end;

function XML.TElement.TChildNodes.ReadCount : Cardinal; inline;
begin
 Result := Length(n);
end;

function XML.TElement.TAttributes.ReadNodes(i : Cardinal) : TAttribute; inline;
begin
 Result := a[i];
end;

function XML.TElement.TAttributes.ReadCount : Cardinal; inline;
begin
 Result := Length(a);
end;

procedure XML.TElement.appendChild(NewChild : TNode);
var i : integer;
begin
 if (NewChild._parentEl <> nil) then NewChild._parentEl.removeChild(NewChild);
 i := Length(childNodes.n);
 SetLength(childNodes.n, i+1);
 childNodes.n[i] := NewChild;
 NewChild._parentEl := self;
end;

procedure XML.TElement.removeChild(i : Cardinal); overload;
begin
 if (i >= childNodes.Count) then exit;
 for i := i + 1 to Length(childNodes.n) - 1 do childNodes.n[i-1] := childNodes.n[i];
 SetLength(childNodes.n, i);
end;

procedure XML.TElement.removeChild(child : TNode); overload;
var
  i,k : integer;
begin
 k := Length(childNodes.n)-1;
 i := 0;
 while (i <= k) do if (childNodes[i] = child) then begin
  for i := i to k-1 do childNodes.n[i] := childNodes.n[i+1];
  SetLength(childNodes.n, k);
  break;
 end else i := i + 1;
end;

function XML.TElement.getAttribute(name : ShortString) : String;
var i : integer;
begin
 Result := '';
 for i := 0 to Length(attributes.a)-1 do if (attributes.a[i].name = name) then begin Result := attributes.a[i].value; break; end;
end;

procedure XML.TElement.setAttribute(name : ShortString; value : String);
var
  P : PAttribute;
  i : integer;
begin
 P := nil;
 for i := 0 to Length(attributes.a)-1 do if (attributes.a[i].name = name) then begin P := @attributes.a[i]; break; end;
 if (value = '') and (P <> nil) then begin
  for i := i to Length(attributes.a) - 2 do attributes.a[i] := attributes[i+1];
  SetLength(attributes.a, Length(attributes.a)-1);
  exit;
 end;
 if (P = nil) then begin
  i := Length(attributes.a);
  SetLength(attributes.a, i+1);
  P := @attributes.a[i];
  P^.name := name;
 end;
 P^.value := value;
end;

function XML.TElement.getElementsByTagName(_tagName : ShortString; i : Cardinal) : TElement;
var j : integer;
begin
 Result := nil;
 for j := 0 to Length(childNodes.n)-1 do if (childNodes.n[j].nodeType = ELEMENT_NODE) and (TElement(childNodes.n[j]).tagName = _tagName) then begin
  if (i = 0) then begin Result := TElement(childNodes.n[j]); exit; end;
  i := i - 1;
 end;
end;

function XML.TElement.getTextContent(RemoveSpaces : Boolean) : TStrings;
function chomp (const s : String) : String;
var
  start, stop : integer;
label SearchStart;
label SearchStop;
begin
 Result := '';
 stop := Length(s);
 SearchStop:
 if (stop <= 1) then exit;
 case Byte(s[stop]) of
  13 : begin stop := stop - 1; goto SearchStop; end;
  10 : begin stop := stop - 1; goto SearchStop; end;
  32 : begin stop := stop - 1; goto SearchStop; end;
 end;
 start := 1;
 SearchStart:
 case Byte(s[start]) of
  13 : begin start := start + 1; goto SearchStart; end;
  10 : begin start := start + 1; goto SearchStart; end;
  32 : begin start := start + 1; goto SearchStart; end;
 end;
 SetLength(Result, stop - start + 1);
 move(s[start],Result[1],stop-start+1);
end;

procedure append (el : TElement);
var
  s    : String;
  i, j : integer;
begin
 for i := 0 to Length(childNodes.n) - 1 do case childNodes[i].nodeType of
  ELEMENT_NODE : append(XML.TElement(childNodes[i]));
  TEXT_NODE : begin
   for j := 0 to XML.TText(childNodes[i]).text.Count - 1 do if (RemoveSpaces) then begin
    s := chomp(XML.TText(childNodes[i]).text[j]);
    if (s <> '') then Result.add(s);
   end else Result.add(XML.TText(childNodes[i]).text[j]);
  end;
 end;
end;

begin
 Result := TStringList.Create;
 append(self);
end;

procedure XML.TElement.clearNodes;
var i : integer;
begin
 for i := Length(childNodes.n) - 1 downto 0 do begin
  childNodes.n[i]._parentEl := nil;
  childNodes.n[i].Free;
 end;
 SetLength(childNodes.n, 0);
end;

destructor XML.TElement.Destroy;
begin
 clearNodes;
 SetLength(attributes.a, 0);
end;



destructor XML.TNode.Destroy;
begin
 if (_parentEl <> nil) then _parentEl.removeChild(self);
end;

end.
